from django.conf import settings
from edc_model.models import BaseUuidModel


class Dashboard(BaseUuidModel):

    # see edc_auth for permissions attached to this model
    # create_edc_dashboard_permissions

    pass


if settings.APP_NAME == "edc_dashboard":
    from .tests import models  # noqa



from .administration_view_mixin import AdministrationViewMixin
from .edc_view_mixin import EdcViewMixin
from .listboard import (
    ListboardFilterViewMixin,
    QueryStringViewMixin,
    SearchFormViewMixin,
    SearchListboardMixin,
    SiteQuerysetViewMixin,
)
from .message_view_mixin import MessageViewMixin
from .template_request_context_mixin import TemplateRequestContextMixin
from .url_request_context_mixin import UrlRequestContextMixin



from django.conf import settings
from edc_constants.constants import (
    ABNORMAL,
    CANCELLED,
    CLOSED,
    COMPLETE,
    FEMALE,
    INCOMPLETE,
    MALE,
    NEW,
    NO,
    NOT_APPLICABLE,
    OPEN,
    OTHER,
    PENDING,
    YES,
)

from .dashboard_templates import dashboard_templates
from .url_names import url_names
from .utils import insert_bootstrap_version


class DashboardMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        try:
            request.url_name_data
        except AttributeError:
            request.url_name_data = url_names.registry
        try:
            request.template_data
        except AttributeError:
            request.template_data = {}
        request.template_data = insert_bootstrap_version(**request.template_data)
        response = self.get_response(request)
        return response

    def process_view(self, request, *args):
        """Adds/Updates references to urls and templates."""
        template_data = dashboard_templates
        try:
            template_data.update(settings.DASHBOARD_BASE_TEMPLATES)
        except AttributeError:
            pass
        template_data = insert_bootstrap_version(**template_data)
        request.template_data.update(**template_data)

    def process_template_response(self, request, response):
        if response.context_data:
            response.context_data.update(
                ABNORMAL=ABNORMAL,
                CANCELLED=CANCELLED,
                CLOSED=CLOSED,
                COMPLETE=COMPLETE,
                DEBUG=settings.DEBUG,
                FEMALE=FEMALE,
                INCOMPLETE=INCOMPLETE,
                MALE=MALE,
                NEW=NEW,
                NO=NO,
                NOT_APPLICABLE=NOT_APPLICABLE,
                OPEN=OPEN,
                OTHER=OTHER,
                PENDING=PENDING,
                SITE_ID=settings.SITE_ID,
                YES=YES,
            )
            if "project_name" not in response.context_data:
                response.context_data.update(project_name="project_name")
        return response



"""To customize any of the values below,
use settings.DASHBOARD_BASE_TEMPLATES.
"""

dashboard_templates = dict(
    edc_base_template="edc_dashboard/base.html",
    edc_listboard_template="edc_dashboard/listboard.html",
)



dashboard_tuples = [
    (
        "edc_dashboard.view_subject_review_listboard",
        "Can view Subject review listboard",
    ),
    ("edc_dashboard.view_enrolment_listboard", "Can view Enrolment listboard"),
    ("edc_dashboard.view_export_dashboard", "Can view Export Dashboard"),
    ("edc_dashboard.view_screening_listboard", "Can view Screening listboard"),
    ("edc_dashboard.view_subject_listboard", "Can view Subject listboard"),
]


def remove_permissions_to_edc_dashboard_model(auth_updater):
    for group in auth_updater.group_model_cls.objects.all():
        auth_updater.remove_permissions_by_codenames(
            group=group,
            codenames=[
                "edc_dashboard.add_dashboard",
                "edc_dashboard.change_dashboard",
                "edc_dashboard.delete_dashboard",
                "edc_dashboard.view_dashboard",
            ],
        )



from edc_auth.auth_objects import CLINIC
from edc_auth.site_auths import site_auths

from .auth_objects import dashboard_tuples, remove_permissions_to_edc_dashboard_model

site_auths.add_post_update_func(remove_permissions_to_edc_dashboard_model)
site_auths.add_custom_permissions_tuples(
    model="edc_dashboard.dashboard", codename_tuples=dashboard_tuples
)

site_auths.update_group(
    "edc_dashboard.view_subject_listboard",
    "edc_dashboard.view_subject_review_listboard",
    name=CLINIC,
)



from django.shortcuts import render
from sentry_sdk import last_event_id


def handler500(request, *args, **argv):
    return render(request, "500.html", {"sentry_event_id": last_event_id()}, status=500)



import six
from django.apps import apps as django_apps
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _
from django.views.generic.list import ListView

from ..view_mixins import (
    QueryStringViewMixin,
    SearchListboardMixin,
    SiteQuerysetViewMixin,
    TemplateRequestContextMixin,
    UrlRequestContextMixin,
)


class ListboardViewError(Exception):
    pass


class BaseListboardView(TemplateRequestContextMixin, ListView):

    cleaned_search_term = None
    context_object_name = "results"
    empty_queryset_message = _("Nothing to display.")
    listboard_template = None  # an existing key in request.context_data
    # if self.listboard_url declared through another mixin.
    listboard_url = None  # an existing key in request.context_data
    listboard_back_url = None

    # default, info, success, danger, warning, etc. See Bootstrap.
    listboard_panel_style = "default"
    listboard_fa_icon = None
    listboard_model = None  # label_lower model name or model class
    listboard_model_manager_name = "_default_manager"
    listboard_panel_title = None
    listboard_instructions = None

    permissions_warning_message = _("You do not have permission to view these data.")
    # e.g. "edc_dashboard.view_subject_listboard"
    listboard_view_permission_codename = None
    # e.g. "edc_dashboard.view_subject_listboard"
    listboard_view_only_my_permission_codename = None

    model_wrapper_cls = None
    ordering = "-created"

    orphans = 3
    paginate_by = 10
    paginator_url = None  # defaults to listboard_url

    def get(self, request, *args, **kwargs):
        if not self.has_view_listboard_perms:
            raise PermissionDenied
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        queryset = context.get(self.context_object_name)
        context_object_name = self.get_context_object_name(queryset)
        wrapped_queryset = self.get_wrapped_queryset(queryset)
        if self.listboard_fa_icon and self.listboard_fa_icon.startswith("fa-"):
            self.listboard_fa_icon = f"fas {self.listboard_fa_icon}"
        context.update(
            empty_queryset_message=self.empty_queryset_message,
            listboard_fa_icon=self.listboard_fa_icon,
            listboard_panel_style=self.listboard_panel_style,
            listboard_panel_title=self.listboard_panel_title,
            listboard_instructions=self.listboard_instructions,
            object_list=wrapped_queryset,
        )
        if context_object_name is not None:
            context[context_object_name] = wrapped_queryset
        context = self.add_url_to_context(
            new_key="listboard_url", existing_key=self.listboard_url, context=context
        )
        if self.listboard_back_url:
            context = self.add_url_to_context(
                new_key="listboard_back_url",
                existing_key=self.listboard_back_url,
                context=context,
            )
        context = self.add_url_to_context(
            new_key="paginator_url",
            existing_key=self.paginator_url or self.listboard_url,
            context=context,
        )

        context.update(
            has_listboard_model_perms=self.has_listboard_model_perms,
            has_view_listboard_perms=self.has_view_listboard_perms,
            listboard_view_permission_codename=self.listboard_view_permission_codename,
            permissions_warning_message=self.permissions_warning_message,
        )

        return context

    def get_template_names(self):
        return [self.get_template_from_context(self.listboard_template)]

    @property
    def url_kwargs(self):
        """Returns a dictionary of URL options for either the
        Search form URL and the Form Action.
        """
        return {}

    @property
    def listboard_model_cls(self):
        """Returns the listboard's model class.

        Accepts `listboard_model` as a model class or label_lower.
        """
        if not self.listboard_model:
            raise ListboardViewError(
                f"Listboard model not declared. Got None. See {repr(self)}"
            )
        try:
            return django_apps.get_model(self.listboard_model)
        except (ValueError, AttributeError):
            return self.listboard_model

    def get_queryset_exclude_options(self, request, *args, **kwargs):
        """Returns exclude options applied to every
        queryset.
        """
        return {}

    def get_queryset_filter_options(self, request, *args, **kwargs):
        """Returns filter options applied to every
        queryset.
        """
        return {}

    def get_filtered_queryset(self, filter_options=None, exclude_options=None):
        """Returns a queryset, called by `get_queryset`.

        This can be overridden but be sure to use the default_manager.
        """
        return (
            getattr(self.listboard_model_cls, self.listboard_model_manager_name)
            .filter(**filter_options)
            .exclude(**exclude_options)
        )

    def get_queryset(self):
        """Return the queryset for this view.

        Completely overrides ListView.get_queryset.

        Note:
            the resulting queryset filtering takes allocated
            permissions into account using Django's permissions
            framework.

            Only returns records if user has dashboard permissions to
            do so. See `has_view_listboard_perms`.

            Limits records to those created by the current user if
            `has_view_only_my_listboard_perms` return True.
            See `has_view_only_my_listboard_perms`.

        Passes filter/exclude criteria to `get_filtered_queryset`.

        Note: The returned queryset is set to self.object_list in
        `get()` just before rendering to response.
        """
        queryset = getattr(self.listboard_model_cls, self.listboard_model_manager_name).none()
        if self.has_view_listboard_perms:
            filter_options = self.get_queryset_filter_options(
                self.request, *self.args, **self.kwargs
            )
            if self.has_view_only_my_listboard_perms:
                filter_options.update(user_created=self.request.user.username)
            exclude_options = self.get_queryset_exclude_options(
                self.request, *self.args, **self.kwargs
            )
            queryset = self.get_filtered_queryset(
                filter_options=filter_options, exclude_options=exclude_options
            )
            queryset = self.get_updated_queryset(queryset)
            ordering = self.get_ordering()
            if ordering:
                if isinstance(ordering, six.string_types):
                    ordering = (ordering,)
                queryset = queryset.order_by(*ordering)
        return queryset

    def get_updated_queryset(self, queryset):
        """Return the queryset for this view.

        Hook for a last chance to modify the queryset
        before ordering.
        """
        return queryset

    def get_wrapped_queryset(self, queryset):
        """Returns a list of wrapped model instances.

        Usually is passed the queryset `object_list` and wraps each
        instance just before passing to the template.
        """
        wrapped_objs = []
        for obj in queryset:
            model_wrapper = self.get_model_wrapper_cls()(model_obj=obj)
            model_wrapper = self.update_wrapped_instance(model_wrapper)
            wrapped_objs.append(model_wrapper)
        return wrapped_objs

    def get_model_wrapper_cls(self):
        return self.model_wrapper_cls

    def update_wrapped_instance(self, model_wrapper):
        """Returns a model_wrapper.

        Hook to add attrs to wrapped model instance.
        """
        return model_wrapper

    @property
    def has_view_listboard_perms(self):
        """Returns True if request.user has permissions to
        view the listboard.

        If False, `get_queryset` returns an empty queryset.
        """
        return self.request.user.has_perms([self.listboard_view_permission_codename])

    @property
    def has_view_only_my_listboard_perms(self):
        """Returns True if request.user ONLY has permissions to
        view records created by request.user on the listboard.
        """
        return self.request.user.has_perm(self.listboard_view_only_my_permission_codename)

    @property
    def has_listboard_model_perms(self):
        """Returns True if request.user has permissions to
        add/change the listboard model.

        Does not affect `get_queryset`.

        Used in templates.
        """
        app_label = self.listboard_model_cls._meta.label_lower.split(".")[0]
        model_name = self.listboard_model_cls._meta.label_lower.split(".")[1]
        return self.request.user.has_perms(
            f"{app_label}.add_{model_name}", f"{app_label}.change_{model_name}"
        )


class ListboardView(
    SiteQuerysetViewMixin,
    QueryStringViewMixin,
    UrlRequestContextMixin,
    SearchListboardMixin,
    BaseListboardView,
):
    urlconfig_getattr = "listboard_urls"

    @classmethod
    def get_urlname(cls):
        return cls.listboard_url



class SiteQuerysetViewMixin:
    def get_queryset_filter_options(self, request, *args, **kwargs):
        options = super().get_queryset_filter_options(request, *args, **kwargs)
        options.update(site__name=request.site.name)
        return options



from django.urls.conf import include, path
from django.views.generic.base import RedirectView

from ..url_config import UrlConfig
from ..views import AdministrationView, ListboardView
from .admin import edc_dashboard_admin

app_name = "edc_dashboard"

subject_listboard_url_config = UrlConfig(
    url_name="listboard_url",
    namespace=app_name,
    view_class=ListboardView,
    label="subject_listboard",
    identifier_label="subject_identifier",
    identifier_pattern="/w+",
)

urlpatterns = subject_listboard_url_config.listboard_urls + [
    path("admin/", edc_dashboard_admin.urls),
    path("edc_auth/", include("edc_auth.urls")),
    path("edc_adverse_event/", include("edc_adverse_event.urls")),
    path("edc_randomization/", include("edc_randomization.urls")),
    path("edc_consent/", include("edc_consent.urls")),
    path("edc_dashboard/", include("edc_dashboard.urls")),
    path("edc_export/", include("edc_export.urls")),
    path("edc_device/", include("edc_device.urls")),
    path("edc_protocol/", include("edc_protocol.urls")),
    path("edc_visit_schedule/", include("edc_visit_schedule.urls")),
    path("administration/", AdministrationView.as_view(), name="administration_url"),
    path("", RedirectView.as_view(url="admin/"), name="home_url"),
]



import urllib


class MetaListboardViewFilters(type):
    def __new__(cls, name, bases, attrs):
        parents = [b for b in bases if isinstance(b, MetaListboardViewFilters)]
        if not parents:
            attrs.update({"filters": []})
            attrs.update({"default_include_filter": None})
            attrs.update({"default_exclude_filter": None})
            return super().__new__(cls, name, bases, attrs)
        filters = []
        default_include_filter = None
        default_exclude_filter = None
        for attrname, obj in attrs.items():
            if not attrname.startswith("_"):
                if isinstance(obj, ListboardFilter):
                    obj.name = attrname
                    if obj.default and not obj.exclude_filter:
                        default_include_filter = obj
                    elif obj.default and obj.exclude_filter:
                        default_exclude_filter = obj
                    filters.append(obj)
        filters.sort(key=lambda x: x.position)
        attrs.update({"filters": filters})
        attrs.update({"default_include_filter": default_include_filter})
        attrs.update({"default_exclude_filter": default_exclude_filter})
        return super().__new__(cls, name, bases, attrs)


class ListboardFilter:
    def __init__(
        self,
        name=None,
        label=None,
        lookup=None,
        exclude_filter=None,
        default=None,
        position=None,
    ):
        self.name = name
        self.label = label or name
        self.position = position or 0
        self.exclude_filter = exclude_filter
        if exclude_filter:
            self.attr = "e"
        else:
            self.attr = "f"
        self.lookup = lookup or {}
        self.default = default

    def __repr__(self):
        return (
            "{0.__class__.__name__}({0.name}, {0.label}, "
            "exclude_filter={0.exclude_filter}, {0.default})".format(self)
        )

    @property
    def querystring(self):
        return urllib.parse.urlencode({self.attr: self.name})

    @property
    def lookup_options(self):
        lookup_options = {}
        for k, v in self.lookup.items():
            try:
                lookup_options.update({k: v()})
            except TypeError:
                lookup_options.update({k: v})
        return lookup_options


class ListboardViewFilters(metaclass=MetaListboardViewFilters):
    @property
    def include_filters(self):
        return [f for f in self.filters if f.attr == "f"]

    @property
    def exclude_filters(self):
        return [f for f in self.filters if f.attr == "e"]



import warnings

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic.base import ContextMixin
from django_revision.views import RevisionMixin
from edc_protocol import Protocol
from edc_sites.view_mixins import SiteViewMixin

from .message_view_mixin import MessageViewMixin
from .template_request_context_mixin import TemplateRequestContextMixin


class EdcViewMixin(
    LoginRequiredMixin,
    MessageViewMixin,
    RevisionMixin,
    SiteViewMixin,
    TemplateRequestContextMixin,
    ContextMixin,
):
    """Adds common template variables and warning messages."""

    edc_protocol_app = "edc_protocol"
    edc_device_app = "edc_device"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        try:
            edc_device_app_config = django_apps.get_app_config(self.edc_device_app)
        except LookupError as e:
            edc_device_app_config = None
            warnings.warn(str(e))

        live_system = getattr(settings, "LIVE_SYSTEM", "TEST")
        sentry_dsn = getattr(settings, "SENTRY_DSN", "sentry_dsn?")
        protocol = Protocol()
        context.update(
            {
                "copyright": getattr(protocol, "copyright", "copyright?"),
                "device_id": getattr(edc_device_app_config, "device_id", "device_id?"),
                "device_role": getattr(edc_device_app_config, "device_role", "device_role?"),
                "disclaimer": getattr(protocol, "disclaimer", "disclaimer?"),
                "institution": getattr(protocol, "institution", "institution?"),
                "license": getattr(protocol, "license", "license?"),
                "project_name": getattr(protocol, "project_name", "project_name?"),
                "project_repo": getattr(protocol, "project_repo", "project_repo?"),
                "live_system": live_system,
                "sentry_dsn": sentry_dsn,
            }
        )

        self.check_for_warning_messages(live_system=live_system)

        return context

    def check_for_warning_messages(self, live_system=None):
        if settings.DEBUG:
            messages.add_message(
                self.request,
                messages.ERROR,
                (
                    "This EDC is running in DEBUG-mode. Use for testing only. "
                    "Do not use this system for production data collection!"
                ),
            )
        elif not settings.DEBUG and not live_system:
            messages.add_message(
                self.request,
                messages.WARNING,
                (
                    "This EDC is for testing only. "
                    "Do not use this system for production data collection!"
                ),
            )
        try:
            if settings.WARNING_MESSAGE:
                messages.add_message(
                    self.request,
                    messages.WARNING,
                    settings.WARNING_MESSAGE,
                    extra_tags="warning",
                )
        except AttributeError:
            pass
        if self.request.user.is_superuser:
            messages.add_message(
                self.request,
                messages.ERROR,
                (
                    "You are using a `superuser` account. The EDC does not operate correctly "
                    "with user acounts that have the `superuser` status. "
                    "Update your user account before continuing."
                ),
            )



from django.db.models import Q
from django.utils.html import escape
from django.utils.text import slugify


class SearchListboardMixin:

    search_fields = ["slug"]

    default_querystring_attrs = "q"
    alternate_search_attr = "subject_identifier"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._search_term = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(search_term=self.search_term)
        return context

    def extra_search_options(self, search_term):
        """Returns a list of search Q() objects that will be added to the
        search criteria (OR) for the search queryset.
        """
        return None

    def clean_search_term(self, search_term):
        return search_term

    @property
    def raw_search_term(self):
        raw_search_term = self.request.GET.get(self.default_querystring_attrs)
        if not raw_search_term:
            raw_search_term = self.kwargs.get(self.alternate_search_attr)
        return raw_search_term

    @property
    def search_term(self):
        if not self._search_term:
            search_term = self.raw_search_term
            if search_term:
                search_term = escape(search_term).strip()
            search_term = self.clean_search_term(search_term)
            self._search_term = search_term
        return self._search_term

    @property
    def search_terms(self):
        return self.search_term.split("+")

    def get_search_filtered_queryset(self, filter_options=None, exclude_options=None):
        q_objs = []
        for search_term in self.search_terms:
            for field in self.search_fields:
                q_objs.append(Q(**{f"{field}__icontains": slugify(search_term)}))
        q_objs.extend(self.extra_search_options(search_term) or [])

        # change to OR
        q_objects = None
        for q_object in q_objs:
            if q_objects:
                q_objects |= q_object
            else:
                q_objects = q_object

        # get queryset
        if q_objects:
            queryset = (
                getattr(self.listboard_model_cls, self.listboard_model_manager_name)
                .filter(q_objects, **filter_options)
                .exclude(**exclude_options)
            )
        else:
            queryset = (
                getattr(self.listboard_model_cls, self.listboard_model_manager_name)
                .filter(**filter_options)
                .exclude(**exclude_options)
            )
        return queryset

    def get_filtered_queryset(self, filter_options=None, exclude_options=None):
        """Override to add conditional logic to filter on search term."""
        if self.search_term and "|" not in self.search_term:
            queryset = self.get_search_filtered_queryset(
                filter_options=filter_options, exclude_options=exclude_options
            )
        else:
            queryset = super().get_filtered_queryset(
                filter_options=filter_options, exclude_options=exclude_options
            )
        ordering = self.get_ordering()
        if ordering:
            if isinstance(ordering, str):
                ordering = (ordering,)
            queryset = queryset.order_by(*ordering)
        return queryset
